extern crate libloading as lib;

#[macro_use]
extern crate lazy_static;
extern crate jni;

#[macro_use]
extern crate bmp;

extern crate regex;
use regex::Regex;

use bmp::{Image, Pixel};

extern crate byteorder;
use byteorder::{NativeEndian, ByteOrder};

use jni::JNIEnv;
use jni::objects::JClass;
use jni::sys::jint;

use std::ffi::{CString, CStr};
use std::ptr;
use std::os::raw::{c_char, c_void, c_int};
use std::slice;

lazy_static! {
    static ref RWL: lib::Library = lib::Library::new("RWL21.dll").expect("Unable to open RWL21.dll");
    static ref RwWriteShape: lib::Symbol<'static, unsafe extern "stdcall" fn(*mut c_char, *mut c_void) -> c_int> = unsafe { RWL.get(b"RwWriteShape\0").expect("Unable to load RwWriteShape") };
    static ref RwForAllClumpsInHierarchyPointer: lib::Symbol<'static, unsafe extern "stdcall" fn(*mut c_void, unsafe extern "stdcall" fn(*mut c_void, *mut c_void) -> *mut c_void, *mut c_void) -> *mut c_void> = unsafe { RWL.get(b"RwForAllClumpsInHierarchyPointer\0").expect("Unable to load RwForAllClumpsInHierarchy") };
    static ref RwForAllPolygonsInClumpPointer: lib::Symbol<'static, unsafe extern "stdcall" fn(*mut c_void, unsafe extern "stdcall" fn(*mut c_void, *mut c_void) -> *mut c_void, *mut c_void) -> *mut c_void> = unsafe { RWL.get(b"RwForAllPolygonsInClumpPointer\0").expect("Unable to load RwForAllPolygonsInClump") };
    static ref RwGetTextureName: lib::Symbol<'static, unsafe extern "stdcall" fn(*mut c_void, *mut c_char, i32) -> *mut c_char> = unsafe { RWL.get(b"RwGetTextureName\0").expect("Unable to load RwGetTextureName") };
    static ref RwGetPolygonTexture: lib::Symbol<'static, unsafe extern "stdcall" fn(*mut c_void) -> *mut c_void> = unsafe { RWL.get(b"RwGetPolygonTexture\0").expect("Unable to load RwGetPolygonTexture") };
    static ref RwGetTextureRaster: lib::Symbol<'static, unsafe extern "stdcall" fn(*mut c_void) -> *mut c_void> = unsafe { RWL.get(b"RwGetTextureRaster\0").expect("Unable to load RwGetTextureRaster") };
    static ref RwGetRasterWidth: lib::Symbol<'static, unsafe extern "stdcall" fn(*mut c_void) -> i32> = unsafe { RWL.get(b"RwGetRasterWidth\0").expect("Unable to load RwGetRasterWidth") };
    static ref RwGetRasterHeight: lib::Symbol<'static, unsafe extern "stdcall" fn(*mut c_void) -> i32> = unsafe { RWL.get(b"RwGetRasterHeight\0").expect("Unable to load RwGetRasterHeight") };
    static ref RwGetRasterStride: lib::Symbol<'static, unsafe extern "stdcall" fn(*mut c_void) -> i32> = unsafe { RWL.get(b"RwGetRasterStride\0").expect("Unable to load RwGetRasterStride") };
    static ref RwGetRasterPixels: lib::Symbol<'static, unsafe extern "stdcall" fn(*mut c_void) -> *mut u8> = unsafe { RWL.get(b"RwGetRasterPixels\0").expect("Unable to load RwGetRasterPixels") };
    static ref RwReleaseRasterPixels: lib::Symbol<'static, unsafe extern "stdcall" fn(*mut c_void, *mut u8) -> *mut c_void> = unsafe { RWL.get(b"RwReleaseRasterPixels\0").expect("Unable to load RwReleaseRasterPixels") };
    static ref RwAddTextureToDict: lib::Symbol<'static, unsafe extern "stdcall" fn(*mut c_char, *mut c_void) -> *mut c_char> = unsafe { RWL.get(b"RwAddTextureToDict\0").expect("Unable to load RwAddTextureToDict") };
    static ref RwGetClumpVertexUV: lib::Symbol<'static, unsafe extern "stdcall" fn(*mut c_void, i32, *mut UV) -> *mut UV> = unsafe { RWL.get(b"RwGetClumpVertexUV\0").expect("Unable to load RwGetClumpVertexUV") };
    static ref RwSetClumpVertexUV: lib::Symbol<'static, unsafe extern "stdcall" fn(*mut c_void, i32, f32, f32) -> *mut c_void> = unsafe { RWL.get(b"RwSetClumpVertexUV\0").expect("Unable to load RwSetClumpVertexUV") };
    static ref RwGetClumpNumVertices: lib::Symbol<'static, unsafe extern "stdcall" fn(*mut c_void) -> c_int> = unsafe { RWL.get(b"RwGetClumpNumVertices\0").expect("Unable to load RwGetClumpNumVertices") };

}

lazy_static! {
    static ref TEX_FIX: Regex = Regex::new("Texture (.*)").unwrap();
}

fn for_all_clumps_in_hierarchy<F: FnMut(*mut c_void)>(clump: *mut c_void, mut closure: F) -> *mut c_void {
    unsafe extern "stdcall" fn trampoline<F: FnMut(*mut c_void)>(clump: *mut c_void, closure_pointer: *mut c_void) -> *mut c_void {
        let closure = unsafe { &mut *(closure_pointer as *mut F) };
        closure(clump);
        clump
    }
    let closure_pointer = &mut closure as *mut _ as *mut c_void;
    unsafe {
        RwForAllClumpsInHierarchyPointer(clump, trampoline::<F>, closure_pointer)
    }
}

fn for_all_polygons_in_clump<F: FnMut(*mut c_void)>(clump: *mut c_void, mut closure: F) -> *mut c_void {
    unsafe extern "stdcall" fn trampoline<F: FnMut(*mut c_void)>(polygon: *mut c_void, closure_pointer: *mut c_void) -> *mut c_void {
        let closure = unsafe { &mut *(closure_pointer as *mut F) };
        closure(polygon);
        polygon
    }
    let closure_pointer = &mut closure as *mut _ as *mut c_void;
    unsafe {
        RwForAllPolygonsInClumpPointer(clump, trampoline::<F>, closure_pointer)
    }
}

fn get_texture_name(texture: *mut c_void) -> Option<String> {
    let mut name_buffer = [0; 1024];
    let rw_result = unsafe { RwGetTextureName(texture, name_buffer.as_mut_ptr() as *mut _, 1024) };
    if rw_result.is_null() {
        return None;
    }
    Some(unsafe {
        CStr::from_ptr(name_buffer.as_ptr() as *const _).to_string_lossy().into_owned()
    })
}

#[repr(C)]
#[derive(Debug, Copy, Clone)]
struct UV {
    u: f32,
    v: f32,
}

impl UV {
    fn new() -> UV {
        UV {
            u: 0.0,
            v: 0.0
        }
    }
}

unsafe extern "stdcall" fn reset_uv(clump: *mut c_void) {
    let num_vertices = RwGetClumpNumVertices(clump);
    for vertex in 1 .. (num_vertices + 1) {
        let mut uv = UV::new();
        RwGetClumpVertexUV(clump, vertex, &mut uv);
        RwSetClumpVertexUV(clump, vertex, uv.u, uv.v);
    }
}

fn rgb565_to_rgb888(pixel: u16) -> (u8, u8, u8) {
    // https://stackoverflow.com/questions/2442576/how-does-one-convert-16-bit-rgb565-to-24-bit-rgb888
    let mut red: u16 = (pixel & 0xf800) >> 11;
    red = (red << 3) | (red >> 2);
    let mut green: u16 = (pixel & 0x07e0) >> 5;
    green = (green << 2) | (green >> 4);
    let mut blue: u16 = pixel & 0x001f;
    blue = (blue << 3) | (blue >> 2);
    (red as u8, green as u8, blue as u8)
}



#[no_mangle]
pub extern "system" fn Java_NET_worlds_scape_Camera_writeShape(env: JNIEnv, class: JClass, room_clump_id: jint, env_clump_id: jint, inf_background_clump_id: jint) -> jint {
    std::fs::DirBuilder::new().recursive(true).create("capture").unwrap();
    unsafe {
        let capture_rwx = CString::new("capture/capture.rwx").unwrap().into_raw();
        let capture_env_rwx = CString::new("capture/capture_env.rwx").unwrap().into_raw();
        let capture_inf_background_rwx = CString::new("capture/capture_inf_background.rwx").unwrap().into_raw();
        for clump in &[room_clump_id as *mut c_void, env_clump_id as *mut c_void, inf_background_clump_id as *mut c_void] {
            for_all_clumps_in_hierarchy(*clump, |clump| {
                reset_uv(clump);
                for_all_polygons_in_clump(clump, |polygon| {
                    let texture = RwGetPolygonTexture(polygon);
                    if !texture.is_null() {
                        let raster = RwGetTextureRaster(texture);
                        let width = RwGetRasterWidth(raster) as usize;
                        let height = RwGetRasterHeight(raster) as usize;
                        let stride = RwGetRasterStride(raster) as usize;
                        let pixels = RwGetRasterPixels(raster);
                        let pixels_slice = slice::from_raw_parts(pixels, (height * stride) as usize);
                        let pixel_chunks = pixels_slice.chunks(2);
                        let mut img = bmp::Image::new(width as u32, height as u32);
                        for (index, pixel) in pixel_chunks.enumerate() {
                            let rgb565 = NativeEndian::read_u16(pixel);
                            let rgb888 = rgb565_to_rgb888(rgb565);
                            img.set_pixel((index % width) as u32, (index / width) as u32, px!(rgb888.0, rgb888.1, rgb888.2));
                        }
                        if let Some(texture_name) = get_texture_name(texture) {
                            //std::fs::File::create(format!("capture/capture_{}.bmp", texture_name.replace('|', "_")));
                            img.save(format!("capture/capture_{}.bmp", texture_name.replace('|', "_"))).unwrap();
                        } else {
                            //std::fs::File::create(format!("capture/capture_{:?}.bmp", texture));
                            let new_texture_name = format!("{:?}", texture);
                            let new_texture_name_raw = CString::new(new_texture_name).unwrap().into_raw();
                            RwAddTextureToDict(new_texture_name_raw, texture);
                            img.save(format!("capture/capture_{:?}.bmp", texture));
                        }
                        drop(pixels_slice);
                        RwReleaseRasterPixels(raster, pixels);
                    }
                });
            });
        }
        let result = RwWriteShape(capture_rwx, room_clump_id as *mut c_void) * RwWriteShape(capture_env_rwx, env_clump_id as *mut c_void) * RwWriteShape(capture_inf_background_rwx, inf_background_clump_id as *mut c_void);
        drop(CString::from_raw(capture_rwx));
        drop(CString::from_raw(capture_env_rwx));
        drop(CString::from_raw(capture_inf_background_rwx));
        for filename in &["capture/capture.rwx", "capture/capture_env.rwx", "capture/capture_inf_background.rwx"] {
            use std::fs::File;
            use std::io::{Read, Write};
            let mut file = File::open(filename).unwrap();
            let mut contents = String::new();
            file.read_to_string(&mut contents).unwrap();
            drop(file);
            let new_contents = TEX_FIX.replace_all(&contents, |captures: &regex::Captures| {
                let texture_name = captures.get(1).unwrap().as_str().replace('|', "_");
                if !texture_name.starts_with("NULL") {
                    format!("Texture capture_{}", texture_name)
                } else {
                    "Texture NULL".to_string()
                }
            });
            let mut file = File::create(filename).unwrap();
            file.write_all(new_contents.as_bytes()).unwrap();
        }
        result
    }
}
